/*
 * #%L
 * EUGene :: EUGene
 * 
 * $Id: ModelPropertiesUtil.java 1031 2011-01-21 17:22:03Z tchemit $
 * $HeadURL: http://svn.nuiton.org/svn/eugene/tags/eugene-2.5/eugene/src/main/java/org/nuiton/eugene/ModelPropertiesUtil.java $
 * %%
 * Copyright (C) 2004 - 2010 CodeLutin
 * %%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation, either version 3 of the 
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Lesser Public License for more details.
 * 
 * You should have received a copy of the GNU General Lesser Public 
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/lgpl-3.0.html>.
 * #L%
 */
package org.nuiton.eugene;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.Map;
import java.util.ServiceLoader;
import java.util.TreeMap;

/**
 * Util class which introduces all the stuff to make usage of stereotype
 * and tag values safe.
 *
 * @author tchemit <chemit@codelutin.com>
 * @since 2.3
 */
public class ModelPropertiesUtil {

    /** Logger. */
    private static final Log log = LogFactory.getLog(ModelPropertiesUtil.class);

    /**
     * Obtain a new store of tag values and stereotypes.
     *
     * @param loader  the classloader to use (if none given will use the one of the current thread)
     * @param verbose a verbose flag
     * @return the aggregate store of all atomic one detected in the class-path
     */
    public static ModelPropertiesProvider newStore(ClassLoader loader,
                                                   boolean verbose) {

        if (loader == null) {

            // use the current thread loader
            loader = Thread.currentThread().getContextClassLoader();
        }

        ModelPropertiesProvider store = new AggregateModelPropertiesProvider(loader);
        store.setVerbose(verbose || log.isDebugEnabled());
        try {
            store.init();
        } catch (Exception e) {
            throw new RuntimeException("Could not init store of tag values and stereotypes", e);
        }
        return store;
    }

    /**
     * Obtain a new store of tag values and stereotypes.
     *
     * @param providers the list of providers to use in the store
     * @param verbose   a verbose flag
     * @return the aggregate store of all atomic one detected in the class-path
     */
    public static ModelPropertiesProvider newStore(Iterable<ModelPropertiesProvider> providers,
                                                   boolean verbose) {

        ModelPropertiesProvider store = new AggregateModelPropertiesProvider(providers);
        store.setVerbose(verbose || log.isDebugEnabled());
        try {
            store.init();
        } catch (Exception e) {
            throw new RuntimeException("Could not init store of tag values and stereotypes", e);
        }
        return store;
    }

    /**
     * The Eugene provider of tag values and stereotypes.
     *
     * @author tchemit <chemit@codelutin.com>
     * @plexus.component role="org.nuiton.eugene.ModelPropertiesUtil$ModelPropertiesProvider" role-hint="eugene"
     * @since 2.3
     */
    public static class EugeneModelPropertiesProvider extends ModelPropertiesProvider {

        @Override
        public void init() throws IllegalAccessException {
            scanStereotypeClass(EugeneStereoTypes.class);
            scanTagValueClass(EugeneTagValues.class);
        }
    }

    /**
     * A aggregate provider to box all the ones registred in the class-path via the
     * {@link ServiceLoader} api.
     *
     * @author tchemit <chemit@codelutin.com>
     * @since 2.3
     */
    public static class AggregateModelPropertiesProvider extends ModelPropertiesProvider {

        protected ClassLoader loader;

        private Iterable<ModelPropertiesProvider> providers;

        public AggregateModelPropertiesProvider(ClassLoader loader) {
            this.loader = loader;
        }

        public AggregateModelPropertiesProvider(Iterable<ModelPropertiesProvider> providers) {
            this.providers = providers;
        }

        @Override
        public void init() throws Exception {

            if (providers == null) {

                providers = ServiceLoader.load(ModelPropertiesProvider.class, loader);
            }

            for (ModelPropertiesProvider provider : providers) {
                if (isVerbose()) {
                    log.info("Will init model properties provider " + provider);
                }
                provider.setVerbose(isVerbose());
                provider.init();
                getStereotypeStore().putAll(provider.getStereotypeStore());
                getTagValueStore().putAll(provider.getTagValueStore());
            }
        }
    }

    /**
     * To define a stereotype.
     * <p/>
     * Place this annotation on the constant defining your stereotype.
     * <p/>
     * <br/>
     * Example for stereotype named mystereotype which can be only put on a attribute :
     * <pre>
     * String STEREOTYPE_MYSTEREOTYPE = "mystereotype";
     * \@StereotypeDefinition(target = ObjectModelAttribute.class)
     * </pre
     *
     * @author tchemit <chemit@codelutin.com>
     * @see EugeneTagValues
     * @since 2.3
     */
    @Target(ElementType.FIELD)
    @Retention(RetentionPolicy.RUNTIME)
    public @interface StereotypeDefinition {

        /**
         * Define the types of object model api which can use this tag value.
         *
         * @return the array of target object model element which can accept the tag value.
         */
        Class<?>[] target();

        /**
         * Obtain the english documentation of the stereotype.
         * 
         * @return the documentation of the stereotype.
         * @since 2.3.1
         */
        String documentation();
    }

    /**
     * To define a tag value.
     * <p/>
     * Place this annotation on the constant defining your tag-value.
     * <p/>
     * Example for tag value named mytagvalut which can be only put on a attribute :
     * <pre>
     * String TAG_VALUE_MYTAGVALUE = "mytagvalue";
     * \@TagValueDefinition(target = ObjectModelAttribute.class)
     * </pre
     *
     * @author tchemit <chemit@codelutin.com>
     * @see EugeneTagValues
     * @since 2.3
     */
    @Target(ElementType.FIELD)
    @Retention(RetentionPolicy.RUNTIME)
    public @interface TagValueDefinition {

        /**
         * Define the types of object model api which can use this tag value.
         *
         * @return the array of target object model element which can accept the tag value.
         */
        Class<?>[] target();

        /**
         * Obtain the english documentation of the tag value.
         *
         * @return the documentation of the tag value.
         * @since 2.3.1
         */
        String documentation();
    }

    /**
     * TODO
     *
     * @author tchemit <chemit@codelutin.com>
     * @since 2.3
     */
    public abstract static class ModelPropertiesProvider {

        /** Logger. */
        private static final Log log =
                LogFactory.getLog(ModelPropertiesProvider.class);

        protected Map<String, TagValueDefinition> tagValueStore;

        protected Map<String, StereotypeDefinition> stereotypeStore;

        protected boolean verbose;

        protected ModelPropertiesProvider() {
            stereotypeStore = new TreeMap<String, StereotypeDefinition>();
            tagValueStore = new TreeMap<String, TagValueDefinition>();
        }

        public boolean isVerbose() {
            return verbose;
        }

        public void setVerbose(boolean verbose) {
            this.verbose = verbose;
        }
        
        public abstract void init() throws Exception;

        protected void scanStereotypeClass(Class<?> holder) throws IllegalAccessException {
            if (isVerbose()) {
                log.info("Will scan " + holder.getName() + " to search some stereotype definitions...");
            }
            Field[] fields = holder.getDeclaredFields();
            for (Field field : fields) {

                StereotypeDefinition stereotypeDefinition =
                        field.getAnnotation(StereotypeDefinition.class);

                if (stereotypeDefinition != null) {
                    String fieldName = field.getName();
                    String stereotypeName = (String) field.get(null);
                    if (isVerbose()) {
                        log.info("Detected stereotype definition [" + fieldName + ":" + stereotypeName + "] : " + Arrays.toString(stereotypeDefinition.target()));
                    }
                    stereotypeStore.put(stereotypeName, stereotypeDefinition);
                }
            }
        }

        protected void scanTagValueClass(Class<?> holder) throws IllegalAccessException {
            if (isVerbose()) {
                log.info("Will scan " + holder.getName() + " to search some tag value definitions...");
            }
            Field[] fields = holder.getDeclaredFields();
            for (Field field : fields) {

                TagValueDefinition tagValueDefinition =
                        field.getAnnotation(TagValueDefinition.class);
                if (tagValueDefinition != null) {
                    String fieldName = field.getName();
                    String tagValueName = (String) field.get(null);
                    if (isVerbose()) {
                        log.info("Detected tag value definition [" + fieldName + ":" + tagValueName + "] : " + Arrays.toString(tagValueDefinition.target()));
                    }
                    tagValueStore.put(tagValueName, tagValueDefinition);
                }
            }
        }

        public Map<String, StereotypeDefinition> getStereotypeStore() {
            return stereotypeStore;
        }

        public Map<String, TagValueDefinition> getTagValueStore() {
            return tagValueStore;
        }

        public Class<?>[] getTagValueTarget(String tagValueName) {
            TagValueDefinition definition = tagValueStore.get(tagValueName);

            Class<?>[] result = null;
            if (definition != null) {
                result = definition.target();
            }
            return result;
        }

        public Class<?>[] getStereotypeTarget(String tagValueName) {

            StereotypeDefinition definition = stereotypeStore.get(tagValueName);

            Class<?>[] result = null;
            if (definition != null) {
                result = definition.target();
            }
            return result;
        }

        public boolean containsTagValue(String tagValueName) {
            boolean b = tagValueStore.containsKey(tagValueName);
            return b;
        }

        public boolean containsStereotype(String stereotypeName) {
            boolean b = stereotypeStore.containsKey(stereotypeName);
            return b;
        }

        public boolean acceptTagValue(String tagValueName, Class<?> type) {
            if (!containsTagValue(tagValueName)) {

                // unknown
                return false;
            }

            Class<?>[] tagValueTarget = getTagValueTarget(tagValueName);

            for (Class<?> target : tagValueTarget) {
                if (target.equals(type) || target.isAssignableFrom(type)) {

                    // found one accepting target
                    return true;
                }
            }

            // no target found
            return false;
        }

        public boolean acceptStereotype(String stereotypeName, Class<?> type) {
            if (!containsStereotype(stereotypeName)) {

                // unknown
                return false;
            }

            Class<?>[] tagValueTarget = getStereotypeTarget(stereotypeName);
            for (Class<?> target : tagValueTarget) {
                if (target.equals(type) || target.isAssignableFrom(type)) {

                    // found one accepting target
                    return true;
                }
            }

            // no target found
            return false;
        }
    }

}
