/*
 * #%L
 * EUGene :: EUGene
 * 
 * $Id: JavaGeneratorUtil.java 1124 2011-12-19 22:23:28Z tchemit $
 * $HeadURL: http://svn.nuiton.org/svn/eugene/tags/eugene-2.5/eugene/src/main/java/org/nuiton/eugene/java/JavaGeneratorUtil.java $
 * %%
 * Copyright (C) 2004 - 2010 CodeLutin
 * %%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation, either version 3 of the 
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Lesser Public License for more details.
 * 
 * You should have received a copy of the GNU General Lesser Public 
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/lgpl-3.0.html>.
 * #L%
 */
package org.nuiton.eugene.java;

import org.apache.commons.lang3.StringUtils;
import org.nuiton.eugene.EugeneStereoTypes;
import org.nuiton.eugene.EugeneTagValues;
import org.nuiton.eugene.GeneratorUtil;
import org.nuiton.eugene.models.object.ObjectModel;
import org.nuiton.eugene.models.object.ObjectModelClassifier;
import org.nuiton.eugene.models.object.ObjectModelElement;
import org.nuiton.eugene.models.object.ObjectModelModifier;
import org.nuiton.eugene.models.object.ObjectModelOperation;
import org.nuiton.eugene.models.object.ObjectModelParameter;

import java.util.ArrayList;
import java.util.List;

/**
 * Utility class for pure java templates.
 *
 * @author tchemit <chemit@codelutin.com>
 * @since 2.0.2
 */
public class JavaGeneratorUtil extends GeneratorUtil {

    public static final String OPERATION_GETTER_DEFAULT_PREFIX = "get";

    public static final String OPERATION_GETTER_BOOLEAN_PREFIX = "is";

    /**
     * dependency to add constants in interfaces via an enumeration.
     * <p/>
     * The literals of enumeration will be the value, and constant names will
     * be generated by transformers.
     *
     * @since 2.0.2
     */
    public static final String DEPENDENCIES_CONSTANTS = "constants";

    /**
     * Duplicates in the {@code target} classifier the given {@code operations}
     * using a {@code transformer} to modify model.
     *
     * @param transformer    the transformer to use
     * @param operations     operations to duplicate
     * @param target         where to duplicate operations
     * @param useVisibility  flag to use operation visibilty to compute his
     *                       modifiers
     * @param extraModifiers scopes to apply to all operations
     */
    public static void cloneOperations(ObjectModelTransformerToJava transformer,
                                       Iterable<ObjectModelOperation> operations,
                                       ObjectModelClassifier target,
                                       boolean useVisibility,
                                       ObjectModelModifier... extraModifiers) {

        for (ObjectModelOperation op : operations) {
            ObjectModelOperation resultOp;
            ObjectModelModifier[] modifiers;
            if (useVisibility) {

                // compute visibility modifer
                String visibility = op.getVisibility();
                ObjectModelModifier modifier =
                        ObjectModelModifier.toValue(visibility);

                int length = extraModifiers.length;
                if (length == 0) {
                    modifiers = new ObjectModelModifier[]{modifier};
                } else {
                    modifiers = new ObjectModelModifier[length + 1];
                    modifiers[0] = modifier;
                    System.arraycopy(extraModifiers, 0, modifiers, 1, length);
                }
            } else {

                // just use the incoming modifiers
                modifiers = extraModifiers;
            }

            resultOp = transformer.addOperation(target,
                                                op.getName(),
                                                op.getReturnType(),
                                                modifiers
            );
            for (ObjectModelParameter param : op.getParameters()) {
                transformer.addParameter(resultOp,
                                         param.getType(),
                                         param.getName()
                );
            }
            for (String exception : op.getExceptions()) {
                transformer.addException(resultOp, exception);
            }
        }
    }

    /**
     * Check if the given classifier has the
     * {@link EugeneStereoTypes#STEREOTYPE_BEAN} stereotype.
     *
     * @param classifier classifier to test
     * @return {@code true} if stereotype was found, {@code false otherwise}
     * @see EugeneStereoTypes#STEREOTYPE_BEAN
     */
    public static boolean hasBeanStereotype(ObjectModelClassifier classifier) {
        return classifier.hasStereotype(EugeneStereoTypes.STEREOTYPE_BEAN);
    }

    /**
     * Check if the given classifier has the
     * {@link EugeneStereoTypes#STEREOTYPE_SKIP} stereotype.
     *
     * @param classifier classifier to test
     * @return {@code true} if stereotype was found, {@code false otherwise}
     * @see EugeneStereoTypes#STEREOTYPE_SKIP
     * @since 2.5
     */
    public static boolean hasSkipStereotype(ObjectModelClassifier classifier) {
        return classifier.hasStereotype(EugeneStereoTypes.STEREOTYPE_SKIP);
    }

    /**
     * Cherche et renvoie le prefixe i18n à utiliser sur cet element, sinon sur
     * le model.
     *
     * @param element l'élément à tester
     * @param model   le modele utilisé
     * @return le prefix i18n ou <code>null</code> si non spécifié
     * @since 2.3
     */
    public static String getI18nPrefixTagValue(ObjectModelElement element,
                                               ObjectModel model) {
        return findTagValue(EugeneTagValues.TAG_I18N_PREFIX, element, model);
    }

    /**
     * Obtain the value of the {@link EugeneTagValues#TAG_DO_NOT_GENERATE_BOOLEAN_GET_METHODS}
     * tag value on the given model or classifier.
     * <p/>
     * It will first look on the model, and then in the given classifier.
     *
     * @param model      model to seek
     * @param classifier classifier to seek
     * @return the none empty value of the found tag value or {@code null} if not found nor empty.
     * @see EugeneTagValues#TAG_DO_NOT_GENERATE_BOOLEAN_GET_METHODS
     * @since 2.4.1
     */
    public static String getDoNotGenerateBooleanGetMethods(ObjectModel model,
                                                           ObjectModelClassifier classifier) {
        String value = findTagValue(EugeneTagValues.TAG_DO_NOT_GENERATE_BOOLEAN_GET_METHODS, classifier, model);
        return value;
    }

    /**
     * Obtain the value of the {@link EugeneTagValues#TAG_NO_PCS}
     * tag value on the given model or classifier.
     * <p/>
     * It will first look on the model, and then in the given classifier.
     *
     * @param model      model to seek
     * @param classifier classifier to seek
     * @return the none empty value of the found tag value or {@code null} if not found nor empty.
     * @see EugeneTagValues#TAG_NO_PCS
     * @since 2.3
     */
    public static String getNoPCSTagValue(ObjectModel model, ObjectModelClassifier classifier) {
        String value = findTagValue(EugeneTagValues.TAG_NO_PCS, classifier, model);
        return value;
    }

    /**
     * Obtain the value of the {@link EugeneTagValues#TAG_CONSTANT_PREFIX}
     * tag value on the given model or classifier.
     * <p/>
     * It will first look on the model, and then in the given classifier.
     *
     * @param model      model to seek
     * @param classifier classifier to seek
     * @return the none empty value of the found tag value or {@code null} if not found nor empty.
     * @see EugeneTagValues#TAG_CONSTANT_PREFIX
     * @since 2.3
     */
    public static String getConstantPrefixTagValue(ObjectModel model,
                                                   ObjectModelClassifier classifier) {
        String value = findTagValue(EugeneTagValues.TAG_CONSTANT_PREFIX, classifier, model);
        return value;
    }

    /**
     * Convertit une propriété javaBean son suffix utilisé pour écrire les
     * getter/setter correspondant.
     *
     * @param propertyName le nom de propriété a convertir
     * @return le nom capitalizé de la propriété JavaBean
     * @since 2.4.2
     */
    public static String capitalizeJavaBeanPropertyName(String propertyName) {

        if (StringUtils.isEmpty(propertyName)) {
            return propertyName;
        }

        if (propertyName.length() == 1) {

            // simple case :
            return propertyName.toUpperCase();
        }

        // check if second caracter is up
        char c = propertyName.charAt(1);
        boolean carIsUp = Character.isUpperCase(c);
        if (carIsUp) {

            // this is a special and strange case : must not capitalize first char
            return propertyName;
        }

        return StringUtils.capitalize(propertyName);
    }

    /**
     * Split the given fqn which contains a generic declaration into his parts.
     * <p/>
     * Example :
     * <pre>
     * java.util.List&lt;Integer&gt; : [java.util.List, Integer]
     * java.util.List&lt;java.util.List&lt;Integer&gt;&gt; : [java.util.List, java.util.List&lt;Integer&gt;]
     * java.util.Map&lt;Integer, Integer&gt; : [java.util.Map, Integer, Integer]
     * java.util.Map&lt;Integer, java.util.List&lt;Integer&gt;&gt; : [java.util.Map, Integer, java.util.List&lt;Integer&gt;]
     * <p/>
     * </pre>
     * <strong>Note:</strong> We only deal with one level of generics here in
     * order to be able easly to rebuild after all the fqn with simple names...
     *
     * @param fqn the fqn to split
     * @return the array of all parts of the fqn.
     * @since 2.3.2
     */
    public static String[] splitGeneric(String fqn) {
        if (fqn == null) {
            return null;
        }

        int firstPartIndex = fqn.indexOf('<');
        if (firstPartIndex == -1) {

            // no generics, just return the fqn
            return new String[]{fqn};
        }

        List<String> parts = new ArrayList<String>();
        parts.add(fqn.substring(0, firstPartIndex));
        String rest = fqn.substring(firstPartIndex + 1,
                                    fqn.lastIndexOf('>'));

        if (containsGenerics(rest) && rest.contains(",")) {

            // there is others generics in generics, worse case...
            int begin = 0;
            int count = 0;
            int max = rest.length();
            for (int i = 0; i < max; i++) {
                char c = rest.charAt(i);
                if (c == '<') {

                    count++;
                } else if (c == '>') {

                    count--;
                } else if (c == ',') {

                    // arrives on a possible split
                    if (count == 0) {

                        // can safely add this part
                        String part = rest.substring(begin, i);
                        parts.add(part.trim());
                        begin = i + 1;
                    }
                }
            }

            // there is a last part to add ?
            if (begin < max) {
                String part = rest.substring(begin, max);
                parts.add(part.trim());
            }

        } else {

            // simple case : each part of the generics has no other generics
            String[] split = rest.split(",");
            for (String part : split) {
                parts.add(part.trim());
            }
        }

        return parts.toArray(new String[parts.size()]);
    }

    /**
     * Join generics parts of a fqn into aparts.
     * <p/>
     * Example :
     * <pre>
     * [java.util.List, Integer] : java.util.List&lt;Integer&gt;
     * [java.util.List, java.util.List&lt;Integer&gt;] : java.util.List&lt;java.util.List&lt;Integer&gt;&gt;
     * [java.util.Map, Integer, Integer] : java.util.Map&lt;Integer, Integer&gt;
     * [java.util.Map, Integer, java.util.List&lt;Integer&gt;] : java.util.Map&lt;Integer, java.util.List&lt;Integer&gt;&gt;
     * </pre>
     *
     * @param genericParts the parts of fqn
     * @return the fqn from his parts
     * @since 2.3.2
     */
    public static String joinGeneric(String... genericParts) {

        if (genericParts == null || genericParts.length == 0) {

            // this case should never happen ?
            return null;
        }

        if (genericParts.length == 1) {

            // in fact, no generics
            return genericParts[0];
        }

        StringBuilder sb = new StringBuilder(genericParts[0]);

        sb.append("<").append(genericParts[1]);

        for (int i = 2, max = genericParts.length; i < max; i++) {
            String genericPart = genericParts[i];
            sb.append(", ").append(genericPart);
        }
        sb.append('>');
        return sb.toString();
    }

    /**
     * Tells if the given fqn contains a generic declaration (says contains a
     * {@code <} caracter).
     *
     * @param fqn the fqn to test
     * @return {@code true} if given fqn contains a generic declaration
     * @since 2.3.2
     */
    public static boolean containsGenerics(String fqn) {
        return fqn.contains("<");
    }

    /**
     * Split the given fqns gieven the list separator.
     * <p/>
     * Example :
     * <pre>
     * Boolean, File : [Boolean, File]
     * Boolean , java.util.List&lt;Integer&gt; : [Boolean, java.util.List&lt;Integer&gt;]
     * <p/>
     * </pre>
     * <p/>
     * <strong>Note:</strong> You can NOT use as separator {@code '&lt;'} nor
     * {@code '&gt;'} nor {@code ' '}.
     *
     * @param fqns      the fqn list to split
     * @param separator the fqn separactor char
     * @return the array of all parts of the fqn.
     * @since 2.3.2
     */
    public static String[] splitFqnList(String fqns, char separator) {
        if (separator == '<' || separator == '>' || separator == ' ') {
            throw new IllegalArgumentException(
                    "Can not use '<' nor '>' for the separator");
        }
        if (fqns == null) {
            return null;
        }

        List<String> parts = new ArrayList<String>();

        // there is others generics in generics, worse case...
        int begin = 0;
        int count = 0;
        int max = fqns.length();
        for (int i = 0; i < max; i++) {
            char c = fqns.charAt(i);
            if (c == '<') {

                count++;
            } else if (c == '>') {

                count--;
            } else if (c == separator) {

                // arrives on a possible split
                if (count == 0) {

                    // can safely add this part
                    String part = fqns.substring(begin, i);
                    parts.add(part.trim());
                    begin = i + 1;
                }
            }
        }

        // there is a last part to add ?
        if (begin < max) {
            String part = fqns.substring(begin, max);
            parts.add(part.trim());
        }

        return parts.toArray(new String[parts.size()]);
    }
}
